from collections.abc import Coroutine
import unittest
from unittest.mock import Mock

from celeryviz.server import Server
from aiohttp.test_utils import AioHTTPTestCase, loop_context, TestClient, TestServer as AioHTTPTestServer


class ServerTestCase(unittest.TestCase):

    def client_test(self, test_func: Coroutine):
        server = Server(Mock())
        with loop_context() as loop:
            client = TestClient(AioHTTPTestServer(server.app, loop=loop))
            
            loop.run_until_complete(client.start_server())

            loop.run_until_complete(test_func(client))

            loop.run_until_complete(client.close())







import logging

import celery
from celery.app import Celery


def attach_log_sender(app: Celery, logger):

    if not hasattr(logger, 'handlers'):
        raise ValueError('logger must support handlers to send logs')
    
    log_level = getattr(logger, 'level', logging.NOTSET)

    handler = TaskLogEmitter(app, level=log_level)
    logger.addHandler(handler)


class TaskLogEmitter(logging.Handler):
    def __init__(self, app: Celery, level=logging.NOTSET):
        self.app = app
        super(TaskLogEmitter, self).__init__(level=level)

    def emit(self, record):
        with self.app.events.default_dispatcher() as d:
            d.send('task-log', uuid=celery.current_task.request.id, msg=record.getMessage(), levelno=record.levelno,
                   pathname=record.pathname, lineno=record.lineno, name=record.name,  exc_info=record.exc_info)



import unittest
from unittest.mock import Mock, patch

from celeryviz.server import Server
from utils import ServerTestCase


class TestServerFeatures(unittest.TestCase):
    def test_recorder(self):
        mock_filename = str(self) + 'mock_filename'

        with patch('celeryviz.server.Recorder') as mock_recorder:
            Server(Mock(), record=True, file=mock_filename)
            mock_recorder.assert_called_once_with(file_name=mock_filename)


class TestServerRunning(ServerTestCase):
    def test_server_running(self):
        async def test(client):
            resp = await client.get('/app/')
            self.assertEqual(resp.status, 200)

        self.client_test(test)



import unittest
from unittest.mock import mock_open, patch

from celeryviz.recorder import Recorder


class TestRecorder(unittest.TestCase):
    def test_recorder(self):
        mock_filename = 'test.txt'
        mock_dict_data = {'key': 'value'}

        with patch('builtins.open', mock_open(read_data='data')) as m:
            recorder = Recorder(file_name=mock_filename)
            recorder.record(mock_dict_data)

            m.assert_called_once_with(mock_filename, 'a')



CLIENT_NAMESPACE = "/client"
SERVER_NAMESPACE = "/server"

SOCKETIO_HOST_PORT = 9095
SOCKETIO_HOST_LOCATION = 'localhost'
SOCKETIO_HOST_URL = 'http://%s:%d' % (SOCKETIO_HOST_LOCATION,
                                      SOCKETIO_HOST_PORT)

SOCKETIO_CLIENT_NAMESPACE_URL = 'http://%s%s' % (
    SOCKETIO_HOST_URL, CLIENT_NAMESPACE)

CELERY_DATA_EVENT = 'celery_events_data'

DEFAULT_LOG_FILE = './log.ndjson'



import threading
import time
import unittest
import sys
from unittest.mock import AsyncMock, Mock, patch
from celery import Celery

from celeryviz.event_receiver import EventListener


class EventHandlerProcess(threading.Thread):
    """
    A separate process for `EventListener` thread to run in.
    Its not possile to kill a thread.
    """
    def __init__(self, app, event_handler, *args, **kwargs):
        self.event_listener = EventListener(app, event_handler, Mock())
        self.event_listener.daemon = True

        super().__init__(*args, **kwargs)

    def run(self):
        self.event_listener.start()

class TestEventListener(unittest.TestCase):

    def setUp(self) -> None:
        self.app = Celery('example_app')
        self.event_handler_process = EventHandlerProcess(
            self.app, AsyncMock(), daemon=True)
        return super().setUp()

    def test_event_listener(self):
        """
        This tests the event handler is added to event loop and is called when celery event is emitted
        """

        with patch('celeryviz.event_receiver.EventListener.onEvent') \
            as mock_event_handler:

            self.event_handler_process.start()

            time.sleep(0.5)
            with self.app.events.default_dispatcher() as dispatcher:
                dispatcher.send('task-received', uuid='1234')

            time.sleep(0.5)
            mock_event_handler.assert_called_once()


import multiprocessing
import threading
import time
import unittest
from unittest.mock import Mock
from celery import Celery

from socketio import Client

from celeryviz.constants import *
from celeryviz.executor import starter


class ClientThread(threading.Thread):
    def __init__(self, on_event):
        super().__init__()
        self.daemon = True
        self.client = Client(logger=False, engineio_logger=False)
        self.client.on(CELERY_DATA_EVENT, handler=on_event,
                  namespace=CLIENT_NAMESPACE)

    def run(self):
        self.client.connect('http://localhost:%s/' %
                       SOCKETIO_HOST_PORT, wait=True, wait_timeout=5)
        self.client.wait()


# This test is not working. It is not able to connect to the server. I am not sure why.
# Keeping it disconnected until its resolved, maybe someone can lok into this.
class TestIntegration():
    """
    This test the backend end to end. It starts the server and client and checks if the client receives the event.

    """

    payload = {"utcoffset": -6, "uuid": "4897c640-a023-4cb8-ae8e-1df4641a3ba1", "name": "basic_task.add", "args": "()", "kwargs": "{}",
               "root_id": "4897c640-a023-4cb8-ae8e-1df4641a3ba1", "parent_id": None, "retries": 0, "eta": None, "expires": None, "timestamp": 1685367125.9317474, "type": "task-received"}

    def setUp(self) -> None:
        self.app = Celery('example_app')
        self.mock_ctx = Mock()
        self.mock_ctx.obj.app = self.app

        self.on_event = Mock()
        self.server_process = multiprocessing.Process(
            target=starter, args=[self.mock_ctx, False, ""], daemon=True)
        self.client_thread = ClientThread(self.on_event)
        return super().setUp()

    def test_integration(self):

        self.server_process.start()

        time.sleep(1)
        self.client_thread.start()
        time.sleep(1)

        with self.app.events.default_dispatcher() as dispatcher:
            dispatcher.send(**self.payload)
        
        time.sleep(1)
        self.assertIsSubDict(self.payload, self.on_event.call_args.args[0])

    def assertIsSubDict(self, subset, superset):
        for key, val in subset.items():
            self.assertEqual(val, superset[key])

    def tearDown(self) -> None:
        self.client_thread.client.disconnect()
        self.server_process.terminate()
        return super().tearDown()



import json


class Recorder:
    records = []

    def __init__(self, file_name: str) -> None:
        self.file_name = file_name

    def record(self, data: dict) -> None:
        with open(self.file_name, 'a') as f:
            f.write(json.dumps(data) + '\n')



import asyncio

from .event_receiver import EventListener
from .server import Server


def starter(ctx, record, file):
    app = ctx.obj.app

    app.control.enable_events()

    server_loop = asyncio.get_event_loop()
    server = Server(server_loop, record=record, file=file)

    event_listener = EventListener(app, server.event_handler, server_loop)
    event_listener.start()

    server.start()



import asyncio
import logging
import threading

from celery.events.receiver import EventReceiver

from .constants import *

logger = logging.getLogger(__name__)


class EventListener(threading.Thread):
    def onEvent(self, event):
        logger.debug("Event received")
        asyncio.run_coroutine_threadsafe(
            self.event_handler(event), self.server_loop)

    def __init__(self, app, event_handler, server_loop):
        super().__init__(daemon=True)
        self.app = app
        self.event_handler = event_handler
        self.server_loop = server_loop

    def run(self):
        handler = {
            'task-failed': self.onEvent,
            'task-received': self.onEvent,
            'task-revoked': self.onEvent,
            'task-started': self.onEvent,
            'task-succeeded': self.onEvent,
            'task-retried': self.onEvent,
            'task-rejected': self.onEvent,
            'task-log': self.onEvent,
        }
        while True:
            try:
                with self.app.connection() as conn:
                    recv = EventReceiver(conn, handlers=handler, app=self.app)

                    recv.capture(limit=None, timeout=None, wakeup=True)

            except (KeyboardInterrupt, SystemExit):
                try:
                    import _thread as thread
                except ImportError:
                    import thread

                thread.interrupt_main()

            except Exception as exc:
                logger.info('Connection error: %r' % (exc, ))



import logging

import click
from celery.bin.base import CeleryCommand

from .constants import DEFAULT_LOG_FILE
from .executor import starter

LOGGING_LEVELS = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
    'CRITICAL': logging.CRITICAL
}

logger = logging.getLogger(__name__)

def set_log_level(log_level):
    logging.basicConfig(level=logging.CRITICAL)
    logger = logging.getLogger(__package__)
    logger.setLevel(LOGGING_LEVELS[log_level])


@click.command(cls=CeleryCommand,
               context_settings={
               'ignore_unknown_options': True})
@click.option('--record',
              default=False,
              is_flag=True)
@click.option('-l', '--log-level',
              default='INFO',
              type=click.Choice(LOGGING_LEVELS.keys(),
                                case_sensitive=False))
@click.option('--file',
              default=DEFAULT_LOG_FILE,
              type=click.Path())
@click.pass_context
def celeryviz(ctx, record, log_level, file):
    set_log_level(log_level)
    starter(ctx, record, file)



from setuptools import setup, find_packages

def get_requirements():
    return open('./requirements.txt').read().splitlines()

classes = """
    Development Status :: 1 - Planning
    Intended Audience :: Developers
    Topic :: System :: Monitoring
    Programming Language :: Python
    Programming Language :: Python :: 3
    Programming Language :: Python :: 3 :: Only
    Programming Language :: Python :: 3.6
    Programming Language :: Python :: 3.7
    Programming Language :: Python :: 3.8
    Programming Language :: Python :: 3.9
    Programming Language :: Python :: Implementation :: CPython
    Programming Language :: Python :: Implementation :: PyPy
    Operating System :: OS Independent
"""
classifiers = [s.strip() for s in classes.split('\n') if s]

setup(
    name='celeryviz',
    description= "A tool for visualising execution of Celery tasks.",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    version='0.0.1',
    classifiers=classifiers,
    author = "Bhavya Peshavaria",
    packages=find_packages(),
    package_dir={'celeryviz': 'celeryviz'},
    package_data={'celeryviz': ['static/*']},
    install_requires=get_requirements(),
    entry_points={
        'celery.commands': [
            'celeryviz = celeryviz.command:celeryviz',
        ],
    },
)


import logging
import os

import socketio
from aiohttp import web

from .constants import *
from .recorder import Recorder

library_path = os.path.dirname(os.path.realpath(__file__))

logger = logging.getLogger(__name__)

banner = """
================================
        🎉 App Launched!
================================
🌐 URL: http://0.0.0.0:%d/app/
================================
""" % SOCKETIO_HOST_PORT


class ClientNapespace(socketio.AsyncNamespace):
    def on_connect(self, sid, environ):
        logger.info("Client connected")

    def on_disconnect(self, sid):
        logger.info("Client disconnected")

    async def on_message(self, sid, data):
        logger.debug('message received with ', data)
        await self.emit('reply', data=data, namespace=SERVER_NAMESPACE)


class ServerNamespace(socketio.AsyncNamespace):
    def on_connect(self, sid, environ):
        logger.info("Server connected")

    def on_disconnect(self, sid):
        logger.info("Server disconnected")

    async def on_message(self, sid, data):
        logger.debug('message received with ', data)
        await self.emit(CELERY_DATA_EVENT, data=data, namespace=CLIENT_NAMESPACE)


async def frontend_app(request):
    return web.FileResponse(f'{library_path}/static/index.html')


class Server:
    sio = socketio.AsyncServer(cors_allowed_origins='*', namespaces=[SERVER_NAMESPACE, CLIENT_NAMESPACE],
                               async_mode='aiohttp')

    def __init__(self, loop, record: bool = False, file: str = DEFAULT_LOG_FILE):
        self.record = record
        self.file = file
        if self.record:
            self.recorder = Recorder(file_name=file)
            logger.info("Recorder enabled")

        self.sio.register_namespace(ServerNamespace(SERVER_NAMESPACE))
        self.sio.register_namespace(ClientNapespace(CLIENT_NAMESPACE))

        self.loop = loop

        self.app = web.Application()

        self.sio.attach(self.app)

        self.app.router.add_get('/app/', frontend_app)
        self.app.router.add_static(
            '/', path=f'{library_path}/static/', name='static')

    def start(self):
        logger.info("\n" + banner)
        web.run_app(self.app, port=SOCKETIO_HOST_PORT, loop=self.loop,
                    print=None)

    async def event_handler(self, data):

        if self.record:
            self.recorder.record(data)

        await self.sio.emit(CELERY_DATA_EVENT, data=data, namespace=CLIENT_NAMESPACE)



import unittest
from unittest.mock import Mock, patch

from celeryviz.executor import starter


class TestStarter(unittest.TestCase):
    def test_starter(self):

        mock_ctx = Mock()
        mock_record = Mock()

        with patch('celeryviz.executor.EventListener') as mock_event_listener,\
                patch('celeryviz.executor.Server') as mock_server:

            starter(mock_ctx, mock_record, "")

            # Check if events are being enabled
            mock_ctx.obj.app.control.enable_events.assert_called_once()

            # Check if server and event listener are being started
            mock_server.assert_called_once()
            mock_server.return_value.start.assert_called_once()
            mock_event_listener.assert_called_once()
            mock_event_listener.return_value.start.assert_called_once()
